@@ -148,6 +148,8 @@ rest_proxy_call_finalize (GObject *object)
   g_free (priv->payload);
   g_free (priv->status_message);
 
+  g_free (priv->url);
+
   G_OBJECT_CLASS (rest_proxy_call_parent_class)->finalize (object);
 }
 
@@ -721,6 +723,12 @@ prepare_message (RestProxyCall *call, GError **error_out)
   priv = GET_PRIVATE (call);
   call_class = REST_PROXY_CALL_GET_CLASS (call);
 
+  /* Emit a warning if the caller is re-using RestProxyCall objects */
+  if (priv->url)
+  {
+    g_warning (G_STRLOC ": re-use of RestProxyCall %p, don't do this", call);
+  }
+
   bound_url =_rest_proxy_get_bound_url (priv->proxy);
 
   if (_rest_proxy_get_binding_required (priv->proxy) && !bound_url)
@@ -759,7 +767,7 @@ prepare_message (RestProxyCall *call, GError **error_out)
     gchar *content_type;
     gsize content_len;
 
-    if (!call_class->serialize_params (call, &content_type, 
+    if (!call_class->serialize_params (call, &content_type,
                                        &content, &content_len, &error))
     {
       g_propagate_error (error_out, error);
@@ -767,7 +775,7 @@ prepare_message (RestProxyCall *call, GError **error_out)
     }
 
     message = soup_message_new (priv->method, priv->url);
-    soup_message_set_request (message, content_type, 
+    soup_message_set_request (message, content_type,
                               SOUP_MEMORY_TAKE, content, content_len);
 
     g_free (content_type);
@@ -829,7 +837,7 @@ prepare_message (RestProxyCall *call, GError **error_out)
 }
 
 /**
- * rest_proxy_call_async:
+ * rest_proxy_call_async: (skip):
  * @call: The #RestProxyCall
  * @callback: a #RestProxyCallAsyncCallback to invoke on completion of the call
  * @weak_object: The #GObject to weakly reference and tie the lifecycle too
@@ -862,14 +870,13 @@ rest_proxy_call_async (RestProxyCall                *call,
 
   if (priv->cur_call_closure)
   {
-    /* FIXME: Use GError here */
-    g_critical (G_STRLOC ": Call already in progress.");
+    g_warning (G_STRLOC ": re-use of RestProxyCall %p, don't do this", call);
     return FALSE;
   }
 
   message = prepare_message (call, error);
   if (message == NULL)
-    goto error;
+    return FALSE;
 
   closure = g_slice_new0 (RestProxyCallAsyncClosure);
   closure->call = g_object_ref (call);
@@ -892,14 +899,116 @@ rest_proxy_call_async (RestProxyCall                *call,
                              message,
                              _call_message_completed_cb,
                              closure);
-  g_free (priv->url);
-  priv->url = NULL;
   return TRUE;
+}
 
-error:
-  g_free (priv->url);
-  priv->url = NULL;
-  return FALSE;
+static void
+_call_message_call_cancelled_cb (GCancellable  *cancellable,
+                                 RestProxyCall *call)
+{
+  rest_proxy_call_cancel (call);
+}
+
+static void
+_call_message_call_completed_cb (SoupSession *session,
+                                 SoupMessage *message,
+                                 gpointer     user_data)
+{
+  GSimpleAsyncResult *result = user_data;
+  RestProxyCall *call;
+  GError *error = NULL;
+
+  call = REST_PROXY_CALL (
+      g_async_result_get_source_object (G_ASYNC_RESULT (result)));
+
+  // FIXME: disconnect cancellable ?
+
+  finish_call (call, message, &error);
+
+  if (error != NULL)
+    g_simple_async_result_take_error (result, error);
+  else
+    g_simple_async_result_set_op_res_gboolean (result, TRUE);
+
+  g_simple_async_result_complete (result);
+
+  g_object_unref (call);
+  g_object_unref (result);
+}
+
+/**
+ * rest_proxy_call_invoke_async:
+ * @call: a #RestProxyCall
+ * @cancellable: (allow-none): an optional #GCancellable that can be used to
+ *   cancel the call, or %NULL
+ * @callback: (scope async): callback to call when the async call is finished
+ * @user_data: (closure): user data for the callback
+ *
+ * A GIO-style version of rest_proxy_call_async().
+ */
+void
+rest_proxy_call_invoke_async (RestProxyCall      *call,
+                              GCancellable       *cancellable,
+                              GAsyncReadyCallback callback,
+                              gpointer            user_data)
+{
+  GSimpleAsyncResult *result;
+  RestProxyCallPrivate *priv;
+  SoupMessage *message;
+  GError *error = NULL;
+
+  g_return_if_fail (REST_IS_PROXY_CALL (call));
+  priv = GET_PRIVATE (call);
+  g_assert (priv->proxy);
+
+  message = prepare_message (call, &error);
+  if (message == NULL)
+    {
+      g_simple_async_report_take_gerror_in_idle (G_OBJECT (call), callback,
+                                                 user_data, error);
+      return;
+    }
+
+  result = g_simple_async_result_new (G_OBJECT (call), callback,
+                                      user_data, rest_proxy_call_invoke_async);
+
+  if (cancellable != NULL)
+    g_signal_connect (cancellable, "cancelled",
+        G_CALLBACK (_call_message_call_cancelled_cb), call);
+
+  _rest_proxy_queue_message (priv->proxy,
+                             message,
+                             _call_message_call_completed_cb,
+                             result);
+}
+
+/**
+ * rest_proxy_call_invoke_finish:
+ * @call: a #RestProxyCall
+ * @result: the result from the #GAsyncReadyCallback
+ * @error: optional #GError
+ *
+ * Returns: %TRUE on success
+ */
+gboolean
+rest_proxy_call_invoke_finish (RestProxyCall *call,
+                             GAsyncResult  *result,
+                             GError       **error)
+{
+  GSimpleAsyncResult *simple;
+
+  g_return_val_if_fail (REST_IS_PROXY_CALL (call), FALSE);
+  g_return_val_if_fail (G_IS_SIMPLE_ASYNC_RESULT (result), FALSE);
+
+  simple = G_SIMPLE_ASYNC_RESULT (result);
+
+  g_return_val_if_fail (g_simple_async_result_is_valid (result,
+        G_OBJECT (call), rest_proxy_call_invoke_async), FALSE);
+
+  if (g_simple_async_result_propagate_error (simple, error))
+    return FALSE;
+
+  return g_simple_async_result_get_op_res_gboolean (simple);
 }
 
 static void
@@ -955,14 +1064,13 @@ rest_proxy_call_continuous (RestProxyCall                    *call,
 
   if (priv->cur_call_closure)
   {
-    /* FIXME: Use GError here */
-    g_critical (G_STRLOC ": Call already in progress.");
+    g_warning (G_STRLOC ": re-use of RestProxyCall %p, don't do this", call);
     return FALSE;
   }
 
   message = prepare_message (call, error);
   if (message == NULL)
-    goto error;
+    return FALSE;
 
   /* Must turn off accumulation */
   soup_message_body_set_accumulate (message->response_body, FALSE);
@@ -993,14 +1101,7 @@ rest_proxy_call_continuous (RestProxyCall                    *call,
                              message,
                              _continuous_call_message_completed_cb,
                              closure);
-  g_free (priv->url);
-  priv->url = NULL;
   return TRUE;
-
-error:
-  g_free (priv->url);
-  priv->url = NULL;
-  return FALSE;
 }
 
 static void
@@ -1099,14 +1200,13 @@ rest_proxy_call_upload (RestProxyCall                *call,
 
   if (priv->cur_call_closure)
   {
-    /* FIXME: Use GError here */
-    g_critical (G_STRLOC ": Call already in progress.");
+    g_warning (G_STRLOC ": re-use of RestProxyCall %p, don't do this", call);
     return FALSE;
   }
 
   message = prepare_message (call, error);
   if (message == NULL)
-    goto error;
+    return FALSE;
 
   closure = g_slice_new0 (RestProxyCallUploadClosure);
   closure->call = g_object_ref (call);
@@ -1135,22 +1235,18 @@ rest_proxy_call_upload (RestProxyCall                *call,
                              message,
                              _upload_call_message_completed_cb,
                              closure);
-  g_free (priv->url);
-  priv->url = NULL;
   return TRUE;
-
-error:
-  g_free (priv->url);
-  priv->url = NULL;
-  return FALSE;
 }
 
 /**
- * rest_proxy_call_cancel:
+ * rest_proxy_call_cancel: (skip):
  * @call: The #RestProxyCall
  *
  * Cancel this call.  It may be too late to not actually send the message, but
  * the callback will not be invoked.
+ *
+ * N.B. this method should only be used with rest_proxy_call_async() and NOT
+ * rest_proxy_call_invoke_async().
  */
 gboolean
 rest_proxy_call_cancel (RestProxyCall *call)
@@ -1296,9 +1392,9 @@ rest_proxy_call_lookup_response_header (RestProxyCall *call,
  * rest_proxy_call_get_response_headers:
  * @call: The #RestProxyCall
  *
- * Returns:  pointer to a hash table of headers. This hash table must not be
- * changed. You should call g_hash_table_unref() when you have finished with
- * it.
+ * Returns: (transfer container): pointer to a hash table of
+ * headers. This hash table must not be changed. You should call
+ * g_hash_table_unref() when you have finished with it.
  */
 GHashTable *
 rest_proxy_call_get_response_headers (RestProxyCall *call)
